/* as_list.c - List handling functions for AS...
*/

#include "as.h"


/* as_add_node - Adds a main node to passed list.
*/

void as_add_node (nd, hd)
	as_node	nd;				/* pointer to node to add						*/
	as_node	hd;				/* head of list to add to						*/
{
	as_node	n;				/* temp node pointer for traversal				*/

	if (hd IS NULL) {
		fprintf (stderr, "AS_ADD_NODE: called with NULL head...\n");
		exit (ERROR);
	}

	for (n = hd; n->next; n = n->next)
		;	/* find last node of list */

	n->next = nd;	/* set pointer to new node */

	nd->prev = n;
	nd->next = NULL;
}

/* as_add_strnode - Adds a string node to passed list.
*/

void as_add_strnode (nd, hd)
	as_str_node	nd;			/* pointer to node to add						*/
	as_str_node	hd;			/* head of list to add to						*/
{
	as_str_node	n;			/* temp node pointer for traversal				*/

	if (hd IS NULL) {
		fprintf (stderr, "AS_ADD_STRNODE: called with NULL head...\n");
		exit (ERROR);
	}

	for (n = hd; n->next; n = n->next)
		;	/* find last node of list */

	n->next = nd;	/* set pointer to new node */

	nd->prev = n;
	nd->next = NULL;
}


/* as_str_in_list - Hunts for passed string from passed list, and returns
					the associated node, if found, else NULL.
*/

as_node as_str_in_list (str, hd, allow_dflt)
	string	str;					/* string to search for					*/
	as_node	hd;						/* head of list							*/
	bool	allow_dflt;				/* TRUE => allow match on "*" now		*/
{
	as_node	n;						/* temp node pointer for traversal		*/

	if ((hd IS NULL) OR (str IS NULL))
		return (NULL);

	for (n = hd; n; n = n->next) {
		if (streq (n->str, "*") AND (NOT allow_dflt))
			continue;	/* don't yet allow match for default response */

		if (NOT fnmatch (n->str, str, FNM_PATHNAME))
			return (n);

#ifdef DEBUGGING_MODE
		fprintf (stdout, "DEBUG: '%s' doesn't match '%s'...\n", str, n->str);
#endif
	}

	return (NULL);
}

/* as_str_in_strlist - Hunts for passed string from passed str list, and
					returns the associated node, if found, else NULL.
*/

as_str_node as_str_in_strlist (str, hd, allow_dflt)
	string		str;				/* string to search for					*/
	as_str_node	hd;					/* head of list							*/
	bool	allow_dflt;				/* TRUE => allow match on "*" now		*/
{
	as_str_node	n;					/* temp node pointer for traversal		*/

	if ((hd IS NULL) OR (str IS NULL))
		return (NULL);

	for (n = hd; n; n = n->next) {
		if (streq (n->str, "*") AND (NOT allow_dflt))
			continue;	/* don't yet allow match for default response */

		if (NOT fnmatch (n->str, str, FNM_PATHNAME))
			return (n);
	}

	return (NULL);
}

/* as_rand_idx - Returns a random index into passed list.
*/

int as_rand_idx (hd)
	as_node	hd;						/* head of list							*/
{
	as_node	n;						/* temp node pointer for traversal		*/
	int		i;						/* loop counter							*/

	if (hd IS NULL)
		return (-1);

	for (n = hd, i = 0; n; n = n->next, i++)
		;	/* determine number of nodes */

	return (rand() % i);
}

/* as_node_at_idx - Finds and returns the node at the given index.
*/

as_node as_node_at_idx (idx, hd)
	int		idx;					/* the index							*/
	as_node	hd;						/* head of the list						*/
{
	as_node	n;						/* temp node pointer for traversal		*/
	int		i;						/* loop counter							*/

	if ((hd IS NULL) OR (idx < 0))
		return (NULL);

	for (n = hd, i = 0; n AND (i ISNT idx); n = n->next, i++)
		;		/* find the node */

	return (n);		/* n will be NULL or right node; either way, return it */
}

/* as_str_node_at_idx - Finds and returns the string node at the given index.
*/

as_str_node as_str_node_at_idx (idx, hd)
	int			idx;				/* the index							*/
	as_str_node	hd;					/* head of the list						*/
{
	as_str_node	n;					/* temp node pointer for traversal		*/
	int			i;					/* loop counter							*/

	if ((hd IS NULL) OR (idx < 0))
		return (NULL);

	for (n = hd, i = 0; n AND (i ISNT idx); n = n->next, i++)
		;		/* find the node */

	return (n);		/* n will be NULL or right node; either way, return it */
}

/* as_list_free - Frees up all memory in passed list.
*/

void as_list_free (hd)
	as_node	hd;						/* head of the list						*/
{
	as_node	n;						/* temp node pointer for traversal		*/
	as_node	n2;						/* temp node pointer for traversal		*/

	if (hd IS NULL)
		return;

	for (n = hd; n; ) {
		if (n->str)
			free (n->str);
		if (n->str_list)
			as_list_free (n->str_list);	/* recursion */

		n2 = n;
		n = n->next;

		free (n2);
	}
}

/* as_strlist_free - Frees up all memory in passed string list.
*/

void as_strlist_free (hd)
	as_str_node	hd;					/* head of the list						*/
{
	as_str_node	n;					/* temp node pointer for traversal		*/
	as_str_node	n2;					/* temp node pointer for traversal		*/

	if (hd IS NULL)
		return;

	for (n = hd; n; ) {
		if (n->str)
			free (n->str);

		n2 = n;
		n = n->next;

		free (n2);
	}
}

/* as_list_find_default - Finds and returns the default case (if any) from
							the passed list.  A default case is one which is
							just the wildcard "*"...
*/

as_node as_list_find_default (hd)
	as_node	hd;						/* head of list							*/
{
	as_node	n;						/* temp node pointer for traversal		*/

	if (hd IS NULL)
		return (NULL);

	for (n = hd; n; n = n->next) {
		if (streq (n->str, "*"))
			return (n);
	}

	return (NULL);	/* no default case */
}


#ifdef DEBUGGING_MODE

/* as_debug_print_list - Prints out contents of passed list to stderr...
*/

void as_debug_print_list (hd, indent_spaces)
	as_node	hd;						/* head of the list						*/
	int		indent_spaces;			/* # of spaces to indent; for recursion	*/
{
	as_node	n;						/* temp node pointer for traversal		*/

	if (hd IS NULL) {
		fprintf (stderr, "%*s*NULL*\n", indent_spaces, " ");
		return;
	}

	for (n = hd; n; n = n->next) {
		fprintf (stderr, "%*sstr: %s\n", indent_spaces, " ", (n->str ? n->str : "*NULL*"));
		fprintf (stderr, "%*sstr_list:\n", indent_spaces, " ");
		as_debug_print_list (n->str_list, indent_spaces + 2);	/* recurse */
	}

	fprintf (stderr, "%*s*EOL*\n", indent_spaces, " ");
}

/* as_debug_print_strlist - Prints out contents of passed strlist to stderr...
*/

void as_debug_print_strlist (hd, indent_spaces)
	as_str_node	hd;					/* head of the list						*/
	int			indent_spaces;		/* # of spaces to indent; for recursion	*/
{
	as_str_node	n;					/* temp node pointer for traversal		*/

	if (hd IS NULL) {
		fprintf (stderr, "%*s*NULL*\n", indent_spaces, " ");
		return;
	}

	for (n = hd; n; n = n->next) {
		fprintf (stderr, "%*sstr: %s\n", indent_spaces, " ", (n->str ? n->str : "*NULL*"));
	}

	fprintf (stderr, "%*s*EOL*\n", indent_spaces, " ");
}

#endif

